from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

from cleo.events.console_events import COMMAND
from poetry.console.commands.installer_command import EnvCommand, InstallerCommand
from poetry.core.factory import Factory as BaseFactory
from poetry.core.pyproject.toml import PyProjectTOML
from poetry.core.version.pep440.version import PEP440Version
from poetry.factory import Factory
from poetry.plugins import ApplicationPlugin
from poetry.poetry import Poetry

from poetry_workspace.commands import loader
from poetry_workspace.commands.workspace.workspace import WorkspaceCommand
from poetry_workspace.workspace import Workspace, is_workspace_pyproject

if TYPE_CHECKING:
    from cleo.events.console_command_event import ConsoleCommandEvent
    from cleo.events.event_dispatcher import EventDispatcher
    from cleo.io.io import IO
    from poetry.console.application import Application


class WorkspacePlugin(ApplicationPlugin):
    def activate(self, application: "Application") -> None:
        loader.register_commands(application.command_loader)

        dispatcher = application.event_dispatcher
        dispatcher.add_listener(COMMAND, self.on_command, priority=1000)

    def on_command(self, event: "ConsoleCommandEvent", _event_name: str, _dispatcher: "EventDispatcher") -> None:
        command = event.command

        monkeypatch_version_parser()

        workspace = find_workspace(command.application, event.io)
        if workspace is None:
            return

        if isinstance(command, EnvCommand):
            monkeypatch_env_manager(workspace)

        if isinstance(command, InstallerCommand):
            set_installer_poetry(command, event.io, workspace)
            monkeypatch_version_solver(workspace)

        if isinstance(command, WorkspaceCommand):
            command.set_workspace(workspace)


def set_installer_poetry(command: InstallerCommand, io: "IO", workspace: Workspace) -> None:
    poetry = Poetry(
        file=BaseFactory.locate(Path.cwd()),
        local_config=workspace.poetry.local_config,
        package=workspace.poetry.package,
        locker=workspace.poetry.locker,
        config=workspace.poetry.config,
    )
    Factory().configure_sources(
        poetry,
        poetry.local_config.get("source", []),
        workspace.poetry.config,
        io,
    )
    command.set_poetry(poetry)


def monkeypatch_env_manager(workspace: Workspace) -> None:
    from poetry.utils.env import EnvManager, SystemEnv, VirtualEnv

    workspace_poetry = workspace.poetry
    original_method = EnvManager.create_venv

    def create_venv(self: EnvManager, *args: Any, **kwargs: Any) -> Union[SystemEnv, VirtualEnv]:
        # Set env manager's Poetry instance to the workspace Poetry instance
        # so that it creates and uses a workspace level virtualenv.
        self._poetry = workspace_poetry
        return original_method(self, *args, **kwargs)

    EnvManager.create_venv = create_venv


def monkeypatch_version_solver(workspace: Workspace) -> None:
    """
    Normally if a workspace project's dependencies are changed, then the workspace
    will resolve the changed dependencies if an update operation was performed for
    the workspace project (e.g. `poetry lock`, `poetry update`, or `poetry update
    foo`). If an update operation was not performed (e.g. `poetry lock --no-update`,
    or `poetry update not-foo`) then the changed dependencies will not be resolved.

    We want to resolve workspace project dependencies all the time as if they are
    direct dependencies of the workspace root project, regardless of whether an
    update was performed or not. This method monkey patches the `VersionSolver` class
    so that all workspace projects are included in the `use_latest` list to ensure
    that the resolver always uses the latest version of workspace projects.
    """
    from poetry.mixology.result import SolverResult
    from poetry.mixology.version_solver import VersionSolver

    original_method = VersionSolver.solve
    workspace_packages = set(project.package.name for project in workspace.projects)

    def solve(self: VersionSolver) -> SolverResult:
        self._use_latest = sorted(set(self._use_latest) | workspace_packages)
        return original_method(self)

    VersionSolver.solve = solve


def monkeypatch_version_parser() -> None:
    """
    Workaround for https://github.com/python-poetry/poetry/issues/4176.
    """
    original_method = PEP440Version.parse

    def parse(cls, value: str):
        if value:
            value = value.rstrip(".")
        return original_method.__func__(cls, value)

    PEP440Version.parse = classmethod(parse)


def find_workspace(application: "Application", io: "IO") -> Optional[Workspace]:
    cwd = Path.cwd()
    for dir_path in [cwd] + list(cwd.parents):
        pyproject = PyProjectTOML(dir_path / "pyproject.toml")
        if not is_workspace_pyproject(pyproject):
            continue

        workspace = Workspace(pyproject, io)
        if workspace.poetry.file.path == application.poetry.file.path:
            # Currently in workspace's root project tree.
            return workspace

        for project in workspace.projects:
            if project.file.path == application.poetry.file.path:
                # Currently in a workspace project's tree.
                return workspace

    return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['poetry_workspace',
 'poetry_workspace.commands',
 'poetry_workspace.commands.workspace',
 'poetry_workspace.schemas',
 'poetry_workspace.vcs']

package_data = \
{'': ['*']}

install_requires = \
['poetry>=1.2.0a2,<2.0.0']

entry_points = \
{'poetry.application.plugin': ['poetry-workspace-plugin = '
                               'poetry_workspace.plugin:WorkspacePlugin']}

setup_kwargs = {
    'name': 'poetry-workspace-plugin2',
    'version': '0.5.4',
    'description': 'Multi project workspace plugin for Poetry',
    'long_description': '# Poetry Workspace Plugin\n\nThis experimental tool is a [Poetry Plugin](https://python-poetry.org/docs/master/plugins) to support workflows in a multi-project repository.\n\n## Installation\n\nMake sure you are using at least Poetry 1.2.0a2. To install this preview release, run:\n\n```shell\ncurl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/install-poetry.py | python - --preview\n```\n\nInstall this plugin:\n\n```shell\npoetry plugin add poetry-workspace-plugin\n```\n\n## Workspace\n\nA workspace is a collection of Poetry projects that share a single environment.\n\n## Example config to place at the root\n```\n[tool.poetry]\nname = "code"\nversion = "0.1.0"\ndescription = "Opendoor Python workspace"\nauthors = ["Developers <developers@opendoor.com>"]\n\n[[tool.poetry.source]]\nname = "pypi-local"\nurl = "https://opendoor.jfrog.io/opendoor/api/pypi/pip/simple"\nsecondary = true\n\n[tool.poetry.workspace]\ninclude = [\n  "lib/**",\n  "workspace/**",\n]\nexclude = [\n  "lib/dev-tools",\n  "lib/legacy",\n  "lib/template/**",\n]\n\n# IMPORTANT: This pyproject.toml file declares dependencies for the shared Python\n# workspace. If your app does not belong to the\n# workspace (i.e. not included in the `include` section above) then do not add your\n# app dependencies here, it will have no effect. Even if your app does belong to\n# the workspace, prefer adding app specific dependencies in your app\'s project. This\n# section is reserved for workspace level constraints.\n[tool.poetry.dependencies]\npython = "~3.9"\nvirtualenv = "^20.10.0"\n\n[tool.poetry.dev-dependencies]\n"opendoor.dev-tools" = {path = "lib/dev-tools", develop = true}\n"opendoor.tools" = {path = "tools", develop = true}\n```\n',
    'author': 'Martin Liu',
    'author_email': 'martin.xs.liu@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
